"""
Selective kernel unit (SK) by Li et al., with support for splitting
tweak by Wightman.

References:
- Li et al. Selective Kernel Networks.
- Wightman. PyTorch Image Models.
"""


__all__ = [
	'SK',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .lin_norm_act import ConvBNAct
from .mlp import MLP
from .pool import global_avg_pool


class SK(nn.Module):
	"""
	Selective kernel unit.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		stride (int): Stride.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		act (T.Callable): Activation function.
		Default is nn.relu.
		n_branches (int): Number of branches.
		Default is 2.
		reduction_factor (int): Reduction factor for the bottleneck
		MLP.
		Default is 8.
		min_reduction_dim (int): Minimum number of channels for the
		hidden layer of the bottleneck MLP.
		Default is 32.
		split (int): Whether to feed each branch a split
		of the input instead of the entire input.
		Default is True.
	"""
	out_dim: T.Optional[int] = None
	stride: int = 1
	groups: T.Union[int, str] = 1
	act: T.Callable = nn.relu
	n_branches: int = 2
	reduction_factor: int = 8
	min_reduction_dim: int = 16
	split: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		if self.split:
			splits = jnp.split(
				ary=input,
				indices_or_sections=self.n_branches,
				axis=-1,
				)
		
		bs, h, w, in_dim = input.shape
		out_dim = self.out_dim or in_dim
		output = []
		for branch_ind in range(self.n_branches):
			output.append(
				ConvBNAct(
					out_dim=out_dim,
					stride=self.stride,
					groups=self.groups,
					dilation=branch_ind+1,
					act=self.act,
					)(splits[branch_ind] if self.split else input, training=training),
				)
		output = jnp.stack(output, axis=-2)

		attention = jnp.sum(output, axis=-2)
		attention = global_avg_pool(attention)
		attention = MLP(
			out_dim=self.n_branches*out_dim,
			hidden_dim=max(out_dim//self.reduction_factor, self.min_reduction_dim),
			act=self.act,
			bias=False,
			bn=True,
			)(attention, training=training)
		attention = jnp.reshape(attention, (bs, 1, 1, self.n_branches, out_dim))
		attention = nn.softmax(attention, axis=-2)

		output = attention*output
		output = jnp.sum(output, axis=-2)
		return output



"""
Efficient channel attention (ECA) by Wang et al.

References:
- Wang et al. ECA-Net: Efficient Channel Attention for Deep Convolutional Neural Networks.
"""


__all__ = [
	'ECA',
	]


import typing as T
from math import log

from flax import linen as nn
from jax import numpy as jnp

from .pool import global_avg_pool


def get_kernel_size(
	in_dim: int,
	beta: int = 1,
	gamma: int = 2,
	) -> int:
	"""
	Calculates the optimal kernel size for ECA.

	Args:
		in_dim (int): Number of input channels.
		beta (int): Beta parameter.
		Default is 1.
		gamma (int): Gamma parameter.
		Default is 2.
	
	Returns (int): Optimal kernel size for ECA given in_dim.
	"""
	t = int((log(in_dim, 2) + beta) / gamma)
	kernel_size = t if t%2 == 1 else t+1
	return kernel_size


class ECA(nn.Module):
	"""
	Efficient channel attention.

	Args:
		beta (int): Beta parameter for calculating the kernel
		size.
		Default is 1.
		gamma (int): Gamma parameter for calculating the kernel
		size.
		Default is 2.
		kernel_size (T.Optional[int]): Kernel size. If None, beta
		and gamma are used to calculate the kernel size. If passed,
		beta and gamma are ignored.
		Default is None.
		pool (T.Callable): Pooling method used for aggregating the
		activations of each channel.
		Default is global_avg_pool.
		gate (T.Callable): Gating function used to normalize the
		attention scores.
		Default is nn.sigmoid.
	"""
	beta: int = 1
	gamma: int = 2
	kernel_size: T.Optional[int] = None
	pool: T.Callable = global_avg_pool
	gate: T.Callable = nn.sigmoid

	@nn.compact
	def __call__(self, input):
		attention = self.pool(input, retain_spatial=False)
		attention = jnp.expand_dims(attention, axis=-2)
		attention = jnp.swapaxes(attention, axis1=-2, axis2=-1)
		
		kernel_size = self.kernel_size or get_kernel_size(input.shape[-1])
		attention = nn.Conv(
			features=1,
			kernel_size=(kernel_size,),
			padding=(kernel_size//2,),
			use_bias=False,
			)(attention)
		attention = self.gate(attention)

		attention = jnp.swapaxes(attention, axis1=-2, axis2=-1)
		attention = jnp.expand_dims(attention, axis=-2)

		return attention*input



"""
MetaFormer block by Yu et al.

References:
- Yu et al. MetaFormer Is Actually What You Need for Vision.
"""


__all__ = [
	'MetaFormerBlock',
	]


import typing as T

from flax import linen as nn

from .acts import gelu
from .layer_scale import LayerScale
from .mlp import TransformerMLP


class MetaFormerBlock(nn.Module):
	"""
	MetaFormer block.

	Args:
		token_mixer (T.Callable): T.Callable returning the token mixer.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion for the 
		hidden layer of the MLP.
		Default is 4.
		act (T.Callable): Activation for the hidden layer of the MLP.
		Default is gelu.
		dw_kernel_size (T.Optional[int]): Kernel size of a depthwise convolution
		applied immediately after the first linear layer of the MLP. If None,
		no depthwise convolution is applied.
		Default is None.
		layer_norm_eps (T.Optional[float]): Epsilon value passed to 
		layer normalization. If None, no normalization is applied.
		Default is 1e-6.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
	"""
	token_mixer: T.Callable
	mlp_hidden_dim_expansion_factor: float = 4.
	act: T.Callable = gelu
	dw_kernel_size: T.Optional[int] = None
	layer_norm_eps: T.Optional[float] = 1e-6
	layer_scale_init_value: T.Optional[float] = None

	@nn.compact
	def __call__(self, input):
		output = nn.LayerNorm(
			epsilon=self.layer_norm_eps,
			)(input) if self.layer_norm_eps else input
		output = self.token_mixer()(output)
		output = LayerScale(
			init_value=self.layer_scale_init_value,
			)(output)
		output = input+output

		output = TransformerMLP(
			hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
			act=self.act,
			dw_kernel_size=self.dw_kernel_size,
			layer_norm_eps=self.layer_norm_eps,
			layer_scale_init_value=self.layer_scale_init_value,
			)(output)

		return output



"""
EfficientNetV2.
"""


import typing as T

from flax import linen as nn

from .. import layers
from .factory import register_configs


class ResConvBNAct(nn.Module):
	"""
	Convolution followed by batch normalization and an activation function,
	plus a residual connection.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		expansion_factor (int): Expansion factor for the
		output dimension.
		Defautl is 1.
		act (T.Callable): Activation function.
		Default is nn.silu.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	stride: int = 1
	expansion_factor: int = 1
	act: T.Callable = nn.silu
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.expansion_factor*self.out_dim,
			stride=self.stride,
			act=self.act,
			tf=self.tf,
			)(input, training=training)
		return input+output if input.shape == output.shape else output


class FusedMBConv(nn.Module):
	"""
	Fused MBConv.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		expansion_factor (int): Expansion factor for the
		width.
		Default is 1.
		act (T.Callable): Activation function.
		Default is nn.silu.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	stride: int = 1
	expansion_factor: int = 1
	act: T.Callable = nn.silu
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.expansion_factor*input.shape[-1],
			stride=self.stride,
			act=self.act,
			tf=self.tf,
			)(input, training=training)
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			kernel_size=1,
			tf=self.tf,
			)(output, training=training)
		return input+output if input.shape == output.shape else output


class DWSeparableConv(nn.Module):
	"""
	Depthwise separable convolution.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		act (T.Callable): Activation function.
		Default is nn.silu.
		res (bool): Whether to apply a residual summation
		if the input and output shapes are identical.
		Default is False.
		se_reduction_factor (T.Optional[int]): Reduction factor for
		squeeze-and-excitation. If None, no squeeze-and-excitation
		is applied.
		Default is 4.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	stride: int = 1
	act: T.Callable = nn.silu
	res: bool = False
	se_reduction_factor: T.Optional[int] = 4
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			stride=self.stride,
			groups='dw',
			act=self.act,
			tf=self.tf,
			)(input, training=training)
		output = layers.SE(
			reduction_factor=self.se_reduction_factor,
			act=self.act,
			)(output) if self.se_reduction_factor else output
		output = layers.ConvBNAct(
			out_dim=self.out_dim,
			kernel_size=1,
			tf=self.tf,
			)(output, training=training)
		return input+output if self.res and input.shape == output.shape else output


class MBConv(nn.Module):
	"""
	MBConv, a.k.a. inverted residual block.

	Args:
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		expansion_factor (int): Expansion factor for
		the inverted bottleneck.
		Default is 1.
		act (T.Callable): Activation function.
		Default is nn.silu.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	out_dim: int
	stride: int = 1
	expansion_factor: int = 1
	act: T.Callable = nn.silu
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.expansion_factor*input.shape[-1],
			kernel_size=1,
			act=self.act,
			tf=self.tf,
			)(input, training=training)
		output = DWSeparableConv(
			out_dim=self.out_dim,
			stride=self.stride,
			act=self.act,
			se_reduction_factor=self.expansion_factor*4,
			tf=self.tf,
			)(output, training=training)
		return input+output if input.shape == output.shape else output


class EfficientNetV2Stage(nn.Module):
	"""
	EfficientNetV2 stage.

	Args:
		block (T.Callable): Block used to construct the stage.
		depth (int): Depth.
		out_dim (int): Number of output channels.
		stride (int): Stride.
		Default is 1.
		expansion_factor (int): Expansion factor.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	block: T.Callable
	depth: int
	out_dim: int
	stride: int = 1
	expansion_factor: int = 1
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		for block_ind in range(self.depth):
			input = self.block(
				out_dim=self.out_dim,
				stride=self.stride if block_ind == 0 else 1,
				expansion_factor=self.expansion_factor,
				tf=self.tf,
				)(input, training=training)
		return input


class EfficientNetV2Head(nn.Module):
	"""
	EfficientNetV2 head.

	Args:
		n_classes (int): Number of output classes. If 0, the input is returned.
		If -1, all stages of the head, other than the final linear layer,
		are applied and the output returned. 
		Default is 0.
		conv_bn_act_dim (int): Number of output channels of
		the pre-pooling convolution-normalization-activation block.
		Default is 1280.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
	"""
	n_classes: int = 0
	conv_bn_act_dim: int = 1280
	tf: bool = True

	@nn.compact
	def __call__(self, input, training: bool = True):
		if self.n_classes == 0:
			return input
		
		output = layers.ConvBNAct(
			out_dim=self.conv_bn_act_dim,
			kernel_size=1,
			act=nn.silu,
			tf=self.tf,
			)(input, training=training)
		output = layers.Head(
			n_classes=self.n_classes,
			)(output)
			
		return output


class EfficientNetV2(nn.Module):
	"""
	EfficientNetV2.

	Args:
		blocks (T.Tuple[T.Callable, ...]): Block used to construct each stage.
		depths (T.Tuple[int, ...]): Depth of each stage.
		out_dims (T.Tuple[int, ...]): Number of output channels of each stage.
		strides (T.Tuple[int, ...]): Stride of each stage.
		expansion_factors (T.Tuple[int, ...]): Expansion factor of
		each stage.
		tf (bool): Whether to use batch normalization epsilon of
		1e-3 and padding of 'same' for compatibility with TensorFlow.
		Default is True.
		n_classes (int): Number of output classes. If 0, there is no head,
		head_bottleneck_dim is ignored, and the raw final features are returned.
		If -1, all stages of the head, other than the final linear layer, are applied
		and the output returned.
		Default is 0.
		head_conv_bn_act_dim (int): Number of output channels of
		the pre-pooling convolution-normalization-activation block.
		Default is 1280.
	"""
	blocks: T.Tuple[T.Callable, ...]
	depths: T.Tuple[int, ...]
	out_dims: T.Tuple[int, ...]
	strides: T.Tuple[int, ...]
	expansion_factors: T.Tuple[int, ...]
	tf: bool = True
	n_classes: int = 0
	head_conv_bn_act_dim: int = 1280 

	@nn.compact
	def __call__(self, input, training: bool = True):
		output = layers.ConvBNAct(
			out_dim=self.out_dims[0],
			stride=2,
			act=nn.silu,
			tf=self.tf,
			)(input, training=training)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		for stage_ind in range(len(self.depths)):
			output = EfficientNetV2Stage(
				block=self.blocks[stage_ind],
				depth=self.depths[stage_ind],
				out_dim=self.out_dims[stage_ind],
				stride=self.strides[stage_ind],
				expansion_factor=self.expansion_factors[stage_ind],
				tf=self.tf,
				)(output, training=training)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = EfficientNetV2Head(
			n_classes=self.n_classes,
			conv_bn_act_dim=self.head_conv_bn_act_dim,
			tf=self.tf,
			)(output, training=training)
		
		return output


@register_configs
def get_efficientnetv2_configs() -> T.Tuple[T.Type[EfficientNetV2], T.Dict]:
	"""
	Gets configurations for all available
	EfficientNetV2 models.

	Returns (T.Tuple[T.Type[EfficientNetV2], T.Dict]): The EfficientNetV2 class and
	configurations of all available models.
	"""
	configs = {
		'efficientnetv2_small': {
			'depths': (2, 4, 4, 6, 9, 15),
			'out_dims': (24, 48, 64, 128, 160, 256),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2),
			'expansion_factors': (1, 4, 4, 4, 6, 6),
			},
		'efficientnetv2_medium': {
			'depths': (3, 5, 5, 7, 14, 18, 5),
			'out_dims': (24, 48, 80, 160, 176, 304, 512),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_large': {
			'depths': (4, 7, 7, 10, 19, 25, 7),
			'out_dims': (32, 64, 96, 192, 224, 384, 640),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_small_in22k': {
			'depths': (2, 4, 4, 6, 9, 15),
			'out_dims': (24, 48, 64, 128, 160, 256),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_medium_in22k': {
			'depths': (3, 5, 5, 7, 14, 18, 5),
			'out_dims': (24, 48, 80, 160, 176, 304, 512),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_large_in22k': {
			'depths': (4, 7, 7, 10, 19, 25, 7),
			'out_dims': (32, 64, 96, 192, 224, 384, 640),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_xlarge_in22k': {
			'depths': (4, 8, 8, 16, 24, 32, 8),
			'out_dims': (32, 64, 96, 192, 256, 512, 640),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_small_in22ft1k': {
			'depths': (2, 4, 4, 6, 9, 15),
			'out_dims': (24, 48, 64, 128, 160, 256),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_medium_in22ft1k': {
			'depths': (3, 5, 5, 7, 14, 18, 5),
			'out_dims': (24, 48, 80, 160, 176, 304, 512),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_large_in22ft1k': {
			'depths': (4, 7, 7, 10, 19, 25, 7),
			'out_dims': (32, 64, 96, 192, 224, 384, 640),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		'efficientnetv2_xlarge_in22ft1k': {
			'depths': (4, 8, 8, 16, 24, 32, 8),
			'out_dims': (32, 64, 96, 192, 256, 512, 640),
			'blocks': (ResConvBNAct, FusedMBConv, FusedMBConv, MBConv, MBConv, MBConv, MBConv),
			'strides': (1, 2, 2, 2, 1, 2, 1),
			'expansion_factors': (1, 4, 4, 4, 6, 6, 6),
			},
		}
	return EfficientNetV2, configs



"""
Split attention (SplAt) by Wang et al.

References:
- Zhang et al. ResNeSt: Split-Attention Networks.
"""


__all__ = [
	'SplAt',
	]


import typing as T

from flax import linen as nn
from jax import numpy as jnp

from .lin_norm_act import ConvBNAct
from .mlp import ConvMLP
from .pool import global_avg_pool


class RadixSoftmax(nn.Module):
	"""
	Softmax along the radix axis.

	Args:
		radix (int): Radix.
		Default is 2.
		groups (int): Number of cardinal groups.
		Default is 1.
	"""
	radix: int = 2
	groups: int = 1

	def __call__(self, input):
		bs, h, w, in_dim = input.shape
		output = jnp.reshape(input, (bs, h, w, self.groups, self.radix, -1))
		output = nn.softmax(output, axis=-2) if 1 < self.radix else nn.sigmoid(output)
		output = jnp.reshape(output, (bs, h, w, self.radix, -1))
		return output


class SplAt(nn.Module):
	"""
	Split attention.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		groups (int): Number of cardinal groups.
		Default is 1.
		act (T.Callable): Activation function.
		Default is nn.relu.
		radix (int): Radix.
		Default is 2.
		reduction_factor (int): Reduction factor for the bottleneck
		MLP.
		Default is 4.
		min_reduction_dim (int): Minimum number of channels for the
		hidden layer of the bottleneck MLP.
		Default is 32.
	"""
	out_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 3
	stride: T.Union[T.Tuple[int, int], int] = 1
	groups: int = 1
	act: T.Callable = nn.relu
	radix: int = 2
	reduction_factor: int = 4
	min_reduction_dim: int = 32

	@nn.compact
	def __call__(self, input, training: bool = True):
		bs, h, w, in_dim = input.shape
		out_dim = self.out_dim or in_dim
		width = self.radix*out_dim

		output = ConvBNAct(
			out_dim=width,
			kernel_size=self.kernel_size,
			stride=self.stride,
			groups=self.radix*self.groups,
			act=self.act,
			)(input, training=training)

		if 1 < self.radix:
			attention = jnp.reshape(output, (bs, h, w, self.radix, -1))
			attention = jnp.sum(attention, axis=-2)
		
		else:
			attention = output

		attention = global_avg_pool(attention)
		attention = ConvMLP(
			out_dim=width,
			hidden_dim=max(output.shape[-1]//self.reduction_factor, self.min_reduction_dim),
			groups=self.groups,
			act=self.act,
			bias_force=True,
			bn=True,
			)(attention, training=training)
		attention = RadixSoftmax(
			radix=self.radix,
			groups=self.groups,
			)(attention)

		output = jnp.reshape(output, (bs, h, w, self.radix, out_dim))
		output = attention*output
		output = jnp.sum(output, axis=-2)
		return output



"""
High-order spatial interaction network (HorNet).
"""


import typing as T
from functools import partial

from flax import linen as nn

from .. import layers
from .factory import register_configs


class HorNetStage(nn.Module):
	"""
	HorNet stage.

	Args:
		depths (int): Depth.
		out_dim (int): Number of output channels.
		order (int): Order of g^n convolution.
		Default is 5.
		scale (float): Scale factor of g^n convolution.
		Default is 1/3.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-6.
		downsample (bool): Whether to downsample.
		Default is False.
	"""
	depth: int
	out_dim: int
	order: int = 5
	scale: float = 1/3
	layer_scale_init_value: T.Optional[float] = 1e-6
	downsample: bool = False

	@nn.compact
	def __call__(self, input):
		if self.downsample:
			input = layers.PatchEmbed(
				token_dim=self.out_dim,
				patch_size=2,
				layer_norm_eps=1e-6,
				norm_first=True,
				flatten=False,
				)(input)

		for _ in range(self.depth):
			input = layers.MetaFormerBlock(
				token_mixer=partial(
					layers.GnConv,
					out_dim=self.out_dim,
					order=self.order,
					scale=self.scale,
					),
				layer_scale_init_value=self.layer_scale_init_value,
				)(input)

		return input


class HorNet(nn.Module):
	"""
	High-order spatial interaction network.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		out_dim_first_stage (int): Number of output channels of the first stage.
		orders (T.Tuple[int, ...]): Order of g^n convolution of each stage.
		Default is (2, 3, 4, 5).
		scale (float): Scale factor of g^n convolution.
		Default is 1/3.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-6.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	out_dim_first_stage: int
	orders: T.Tuple[int, ...] = (2, 3, 4, 5)
	scale: float = 1/3
	layer_scale_init_value: T.Optional[float] = 1e-6
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.out_dim_first_stage,
			patch_size=4,
			layer_norm_eps=1e-6,
			flatten=False,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			)
		
		for stage_ind in range(len(self.depths)):
			output = HorNetStage(
				depth=self.depths[stage_ind],
				out_dim=self.out_dim_first_stage * (2**stage_ind),
				order=self.orders[stage_ind],
				scale=self.scale,
				layer_scale_init_value=self.layer_scale_init_value,
				downsample=False if stage_ind == 0 else True,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			layer_norm_eps=1e-6,
			)(output)

		return output


@register_configs
def get_hornet_configs() -> T.Tuple[T.Type[HorNet], T.Dict]:
	"""
	Gets configurations for all available
	HorNet models.

	Returns (T.Tuple[T.Type[HorNet], T.Dict]): HorNet class and
	configurations of all models.
	"""
	configs = {
		'hornet_tiny': {
			'depths': (2, 3, 18, 2),
			'out_dim_first_stage': 64,
			}, 
		'hornet_small': {
			'depths': (2, 3, 18, 2),
			'out_dim_first_stage': 96,
			},
		'hornet_base': {
			'depths': (2, 3, 18, 2),
			'out_dim_first_stage': 128,
			}, 
		'hornet_large_in22k': {
			'depths': (2, 3, 18, 2),
			'out_dim_first_stage': 192,
			},
		}
	return HorNet, configs



"""
Class attention image transformer (Cait).
"""


import typing as T
from functools import partial

from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


class TalkingHeads(nn.Module):
	"""
	Module enabling communication between different heads,
	a.k.a. talking heads.
	"""
	@nn.compact
	def __call__(self, input):
		output = jnp.transpose(input, (0, 2, 3, 1))
		output = nn.Dense(
			features=output.shape[-1],
			)(output)
		output = jnp.transpose(output, (0, 3, 1, 2))
		return output


class ClassAttentionQKV(nn.Module):
	"""
	Query-key-value extractor for class attention.

	Args:
		n_heads (int): Number of heads.
	"""
	n_heads: int

	@nn.compact
	def __call__(self, input):
		n_tokens, token_dim = input.shape[-2:]
		head_dim = token_dim//self.n_heads
		
		q = nn.Dense(
			features=token_dim,
			)(input[:, 0])
		kv = nn.Dense(
			features=2*token_dim,
			)(input)

		q = jnp.reshape(q, (-1, 1, self.n_heads, head_dim))
		kv = jnp.reshape(kv, (-1, n_tokens, 2, self.n_heads, head_dim))

		q = jnp.transpose(q, (0, 2, 1, 3))
		kv = jnp.transpose(kv, (2, 0, 3, 1, 4))
		k, v = jnp.split(
			ary=kv,
			indices_or_sections=2,
			axis=0,
			)

		return q, jnp.squeeze(k, axis=0), jnp.squeeze(v, axis=0)


class ClassAttentionBlock(nn.Module):
	"""
	Transformer block with class attention.

	Args:
		n_heads (int): Number of heads.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-5.
	"""
	n_heads: int
	layer_scale_init_value: T.Optional[float] = 1e-5

	@nn.compact
	def __call__(self, input, class_token):
		concat = jnp.concatenate((class_token, input), axis=1)
		output = nn.LayerNorm()(concat)
		output = layers.MHSA(
			to_qkv=partial(ClassAttentionQKV, n_heads=self.n_heads),
			)(output)
		output = layers.LayerScale(
			init_value=self.layer_scale_init_value,
			)(output)
		class_token = class_token+output

		class_token = layers.TransformerMLP(
			layer_scale_init_value=self.layer_scale_init_value,
			)(class_token)

		return class_token


class Cait(nn.Module):
	"""
	Class-attention image transformer.

	Args:
		depth (int): Depth of the no-class-token 
		part of the model.
		token_dim (int): Token dimension.
		n_heads (int): Number of heads.
		patch_size (int): Patch size. This value
		is used along both spatial dimensions.
		Default is 16.
		depth_class (int): Depth of the class-token-only
		part of the model.
		Default is 2.
		layer_scale_init_value (T.Optional[float]): Value for initializing
		LayerScale. If None, no LayerScale is applied.
		Default is 1e-5.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depth: int
	token_dim: int
	n_heads: int
	patch_size: int = 16
	depth_class: int = 2
	layer_scale_init_value: T.Optional[float] = 1e-5
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = layers.PatchEmbed(
			token_dim=self.token_dim,
			patch_size=self.patch_size,
			)(input)
		output = layers.AbsPosEmbed()(output)
		self.sow(
			col='intermediates',
			name='block_0',
			value=output,
			)

		for block_ind in range(self.depth):
			output = layers.MetaFormerBlock(
				token_mixer=partial(
					layers.MHSA,
					to_qkv=self.n_heads,
					pre_softmax=TalkingHeads,
					post_softmax=TalkingHeads,
					),
				layer_scale_init_value=self.layer_scale_init_value,
				)(output)
			self.sow(
				col='intermediates',
				name=f'block_{block_ind+1}',
				value=output,
				)

		class_token = layers.ClassToken(concat=False)(output)
		for block_ind in range(self.depth_class):
			class_token = ClassAttentionBlock(
				n_heads=self.n_heads,
				layer_scale_init_value=self.layer_scale_init_value,
				)(output, class_token)
			self.sow(
				col='intermediates',
				name=f'block_{self.depth+block_ind+1}',
				value=class_token,
				)
		
		output = jnp.concatenate((class_token, output), axis=1)
		output = layers.ViTHead(
			n_classes=self.n_classes,
			layer_norm_eps=1e-6,
			)(output)
			
		return output


@register_configs
def get_cait_configs() -> T.Tuple[T.Type[Cait], T.Dict]:
	"""
	Gets configurations for all available
	Cait models.

	Returns (T.Tuple[T.Type[Cait], T.Dict]): The Cait class and
	configurations of all available models.
	"""
	configs = {
		'cait_xxsmall24_224': {
			'token_dim': 192,
			'n_heads': 4,
			'depth': 24,
			},
		'cait_xxsmall36_224': {
			'token_dim': 192,
			'n_heads': 4,
			'depth': 36,
			},
		'cait_small24_224': {
			'token_dim': 384,
			'n_heads': 8,
			'depth': 24,
			},
		'cait_xxsmall24_384': {
			'token_dim': 192,
			'n_heads': 4,
			'depth': 24,
			},
		'cait_xxsmall36_384': {
			'token_dim': 192,
			'n_heads': 4,
			'depth': 36,
			},
		'cait_xsmall24_384': {
			'token_dim': 288,
			'n_heads': 6,
			'depth': 24,
			},
		'cait_small24_384': {
			'token_dim': 384,
			'n_heads': 8,
			'depth': 24,
			},
		'cait_small36_384': {
			'token_dim': 384,
			'n_heads': 8,
			'depth': 36,
			'layer_scale_init_value': 1e-6,
			},
		'cait_medium36_384': {
			'token_dim': 768,
			'n_heads': 16,
			'layer_scale_init_value': 1e-6,
			'depth': 36,
			},
		'cait_medium48_448': {
			'token_dim': 768,
			'n_heads': 16,
			'layer_scale_init_value': 1e-6,
			'depth': 48,
			},
		}
	return Cait, configs



"""
MLP-like layers:
- MLP: Multilayer perceptron with one hidden layer.
- ConvMLP: Multilayer perceptron with one hidden layer, implemented using convolutions.
- TransformerMLP: Transformer-style MLP by Vaswani et al.

References:
- Vaswani et al. Attention Is All You Need.
"""


__all__ = [
	'MLP',
	'ConvMLP',
	'TransformerMLP',
	]


import typing as T
from math import sqrt

from flax import linen as nn
from jax import numpy as jnp

from .acts import gelu
from .layer_scale import LayerScale
from .lin_norm_act import Conv


def get_dims(
	in_dim: int,
	out_dim: T.Optional[int],
	hidden_dim_expansion_factor: T.Optional[float],
	hidden_dim: T.Optional[int],
	) -> T.Tuple[int, int]:
	"""
	Gets hidden and output dimensions for an MLP.

	Args:
		in_dim (int): Number of input neurons.
		out_dim (T.Optional[int]): Number of output neurons. If None,
		it is set to the number of input channels.
		Default is None.
		hidden_dim_expansion_factor (T.Optional[float]): Factor of expansion for the 
		hidden layer. If None, hidden_dim must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		hidden_dim (T.Optional[int]): Number of neurons in the hidden layer.
		If None, hidden_dim_expansion_factor must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
	
	Returns (T.Tuple[int, int]): Hidden and output dimensions.
	"""
	return (
		hidden_dim or int(hidden_dim_expansion_factor*in_dim),
		out_dim or in_dim,
		)


class MLP(nn.Module):
	"""
	Multilayer perceptron with one hidden layer.

	Args:
		out_dim (T.Optional[int]): Number of output neurons. If None,
		it is set to the number of input channels.
		Default is None.
		hidden_dim_expansion_factor (T.Optional[float]): Factor of expansion for the 
		hidden layer. If None, hidden_dim must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		hidden_dim (T.Optional[int]): Number of neurons in the hidden layer.
		If None, hidden_dim_expansion_factor must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		act (T.Callable): Activation for the hidden layer.
		Default is nn.relu.
		output_act (bool): Whether to apply activation on the output.
		Default is False. 
		bias (bool): Whether the linear layers should have a bias term. If bn is True,
		the first linear layer would not have a bias term.
		Default is True.
		bias_force (bool): Whether to force the first linear layer to have a bias term 
		even if bn is True.
		Default is False.
		bn (bool): Whether to apply batch normalization before the activation.
		If False, the training argument is ignored.
		Default is False.
		dw_kernel_size (T.Optional[int]): Kernel size of a depthwise convolution
		applied immediately after the first linear layer. If None,
		no depthwise convolution is applied.
		Default is None.
	"""
	out_dim: T.Optional[int] = None
	hidden_dim_expansion_factor: T.Optional[float] = None
	hidden_dim: T.Optional[int] = None
	act: T.Callable = nn.relu
	output_act: bool = False
	bias: bool = True
	bias_force: bool = False
	bn: bool = False
	dw_kernel_size: T.Optional[int] = None

	@nn.compact
	def __call__(self, input, training: bool = True):
		reshape = False
		if self.dw_kernel_size and input.ndim == 3:
			bs, n_tokens, in_dim = input.shape
			img_size = int(sqrt(n_tokens))
			input = jnp.reshape(input, (bs, img_size, img_size, in_dim))
			reshape = True

		hidden_dim, out_dim = get_dims(
			in_dim=input.shape[-1],
			hidden_dim_expansion_factor=self.hidden_dim_expansion_factor,
			hidden_dim=self.hidden_dim,
			out_dim=self.out_dim,
			)
		output = nn.Dense(
			features=hidden_dim,
			use_bias=self.bias_force or (self.bias and not self.bn),
			)(input)
		output = Conv(
			kernel_size=self.dw_kernel_size,
			groups='dw',
			bias=self.bias_force or (self.bias and not self.bn),
			)(output) if self.dw_kernel_size else output
		output = nn.BatchNorm(
			use_running_average=not training,
			)(output) if self.bn else output
		output = self.act(output)
		output = nn.Dense(
			features=out_dim,
			use_bias=self.bias,
			)(output)
		output = self.act(output) if self.output_act else output

		output = jnp.reshape(output, (bs, n_tokens, in_dim)) if reshape else output
		return output


class ConvMLP(nn.Module):
	"""
	Multilayer perceptron with one hidden layer, 
	implenented using convolutions.

	Args:
		out_dim (T.Optional[int]): Number of output channels. If None,
		it is set to the number of input channels.
		Default is None.
		hidden_dim_expansion_factor (T.Optional[float]): Factor of expansion for the 
		hidden layer. If None, hidden_dim must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		hidden_dim (T.Optional[int]): Number of channels in the hidden layer.
		If None, hidden_dim_expansion_factor must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		kernel_size (T.Union[T.Tuple[int, int], int]): Kernel size.
		If an int, this value is used along both spatial dimensions.
		Default is 3.
		stride (T.Union[T.Tuple[int, int], int]): Stride. If an int,
		this value is used along both spatial dimensions.
		Default is 1.
		padding (T.Optional[T.Union[str, int]]): Padding. If None,
		it is set so the spatial dimensions are exactly divided by stride.
		If an int, this value is used along both spatial dimensions.
		Default is None.
		groups (T.Union[int, str]): Number of groups. If 'dw', a depthwise convolution is performed.
		Default is 1.
		dilation (int): Dilation.
		Default is 1.
		act (T.Callable): Activation for the hidden layer.
		Default is nn.relu.
		output_act (bool): Whether to apply activation on the output.
		Default is False. 
		bias (bool): Whether the convolutions should have a bias term. If bn is True,
		the first convolution would not have a bias term.
		Default is True.
		bias_force (bool): Whether to force the first convolution to have a bias term 
		even if bn is True.
		Default is False.
		bn (bool): Whether to apply batch normalization before the activation.
		If False, the training argument is ignored.
		Default is False.
	"""
	out_dim: T.Optional[int] = None
	hidden_dim_expansion_factor: T.Optional[float] = None
	hidden_dim: T.Optional[int] = None
	kernel_size: T.Union[T.Tuple[int, int], int] = 1
	stride: T.Union[T.Tuple[int, int], int] = 1
	padding: T.Optional[T.Union[str, int]] = None
	groups: T.Union[int, str] = 1
	dilation: int = 1
	act: T.Callable = nn.relu
	output_act: bool = False
	bias: bool = True
	bias_force: bool = True
	bn: bool = False

	@nn.compact
	def __call__(self, input, training: bool = True):
		hidden_dim, out_dim = get_dims(
			in_dim=input.shape[-1],
			hidden_dim_expansion_factor=self.hidden_dim_expansion_factor,
			hidden_dim=self.hidden_dim,
			out_dim=self.out_dim,
			)
		output = Conv(
			out_dim=hidden_dim,
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding=self.padding,
			groups=self.groups,
			dilation=self.dilation,
			bias=self.bias_force or (self.bias and not self.bn),
			)(input)
		output = nn.BatchNorm(
			use_running_average=not training,
			)(output) if self.bn else output
		output = self.act(output)
		output = Conv(
			out_dim=out_dim,
			kernel_size=self.kernel_size,
			stride=self.stride,
			padding=self.padding,
			groups=self.groups,
			dilation=self.dilation,
			bias=self.bias_force or self.bias,
			)(output)
		output = self.act(output) if self.output_act else output
		return output


class TransformerMLP(nn.Module):
	"""
	Transformer-style MLP with layer normalization, LayerScale,
	and a residual connection.

	Args:
		out_dim (T.Optional[int]): Number of output neurons. If None,
		it is set to the number of input channels.
		Default is None.
		hidden_dim_expansion_factor (T.Optional[float]): Factor of expansion for the 
		hidden layer. If None, hidden_dim must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		hidden_dim (T.Optional[int]): Number of neurons in the hidden layer.
		If None, hidden_dim_expansion_factor must be provided. If both
		hidden_dim_expansion_factor and hidden_dim are provided, the latter takes precedence.
		Default is None.
		act (T.Callable): Activation for the hidden layer.
		Default is gelu.
		dw_kernel_size (T.Optional[int]): Kernel size of a depthwise convolution
		applied immediately after the first linear layer. If None,
		no depthwise convolution is applied.
		Default is None.
		layer_norm_eps (T.Optional[float]): Epsilon value passed to 
		layer normalization. If None, no normalization is applied.
		Default is 1e-6.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
		residual (bool): Whether a residual summation should be applied
		if the input and output shapes are identical.
		Default is True.
	"""
	out_dim: T.Optional[int] = None
	hidden_dim_expansion_factor: T.Optional[float] = 4
	hidden_dim: T.Optional[int] = None
	act: T.Callable = gelu
	dw_kernel_size: T.Optional[int] = None
	layer_norm_eps: T.Optional[float] = 1e-6
	layer_scale_init_value: T.Optional[float] = None
	residual: bool = True

	@nn.compact
	def __call__(self, input):
		output = nn.LayerNorm(
			epsilon=self.layer_norm_eps,
			)(input) if self.layer_norm_eps else input
		output = MLP(
			out_dim=self.out_dim,
			hidden_dim_expansion_factor=self.hidden_dim_expansion_factor,
			hidden_dim=self.hidden_dim,
			act=self.act,
			dw_kernel_size=self.dw_kernel_size,
			)(output)
		output = LayerScale(
			init_value=self.layer_scale_init_value,
			)(output)
		return input+output if self.residual and input.shape == output.shape else output



"""
Activation functions:
- gelu: Identical to Flax's GeLU but approximate is fixed at False.
- mish: Mish by Misra.
- square_relu: Squared ReLU by So et al.
- StarReLU: StarReLU by Yu et al.

References: 
- Misra. Mish: A Self Regularized Non-Monotonic Activation Function.
- So et al. Primer: Searching for Efficient Transformers for Language Modeling.
- Yu et al. MetaFormer Baselines for Vision.
"""


__all__ = [
	'gelu',
	'mish',
	'squared_relu',
	'StarReLU',
	]


from functools import partial

from flax import linen as nn
from jax import numpy as jnp


gelu = partial(nn.gelu, approximate=False)


def mish(input):
	"""
	Mish activation function.

	Args:
		input: Input.
	
	Returns: Result of mish.
	"""
	return input * jnp.tanh(nn.softplus(input))


def squared_relu(input):
	"""
	Squared ReLU activation function.

	Args:
		input: Input.
	
	Returns: Result of squared ReLU.
	"""
	return nn.relu(input) ** 2


class StarReLU(nn.Module):
	"""
	StarReLU activation function.

	Args:
		scale_init_value (float): Initial value for the
		scale parameter.
		Default is 1.
		bias_init_value (float): Initial value for the
		bias parameter.
		Default is 0.
	"""
	scale_init_value: float = 1.0
	bias_init_value: float = 0.0
	
	@nn.compact
	def __call__(self, input):
		scale = self.param(
			name='scale',
			init_fn=lambda prng: self.scale_init_value,
			)
		bias = self.param(
			name='bias',
			init_fn=lambda prng: self.bias_init_value,
			)
		return scale * squared_relu(input) + bias



"""
Global context vision transformer (GC ViT).
"""


import typing as T
from functools import partial
from math import log2

from flax import linen as nn
from jax import numpy as jnp

from .. import layers
from .factory import register_configs


class GCViTFusedMBConv(nn.Module):
	"""
	Modified version of fused MBConv for GC ViT.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
		expansion_factor (int): Expansion factor for the
		width.
		Default is 1.
		act (T.Callable): Activation function.
		Default is layers.gelu.
	"""
	out_dim: T.Optional[int] = None
	expansion_factor: int = 1
	act: T.Callable = layers.gelu

	@nn.compact
	def __call__(self, input):
		in_dim = input.shape[-1]
		output = layers.ConvBNAct(
			out_dim=self.expansion_factor*in_dim,
			groups='dw',
			bias=False,
			bn=False,
			act=self.act,
			)(input)
		output = layers.SE(
			reduction_factor=4,
			act=layers.gelu,
			bias=False,
			)(output)
		output = layers.Conv(
			out_dim=self.out_dim or in_dim,
			kernel_size=1,
			bias=False,
			)(output)
		return input+output if input.shape == output.shape else output


class GCViTDownsample(nn.Module):
	"""
	GC ViT downsampling module.

	Args:
		out_dim (T.Optional[int]): Number of output channels.
		If None, it is set to the number of input channels.
		Default is None.
	"""
	out_dim: T.Optional[int] = None

	@nn.compact
	def __call__(self, input):
		output = nn.LayerNorm(
			epsilon=1e-5,
			)(input)
		output = GCViTFusedMBConv()(output)
		output = layers.Conv(
			out_dim=self.out_dim,
			stride=2,
			bias=False,
			)(output)
		output = nn.LayerNorm(
			epsilon=1e-5,
			)(output)
		return output

	
class GCViTStem(nn.Module):
	"""
	GC ViT stem.

	Args:
		out_dim (int): Number of output channels.
		Default is 96.
	"""
	out_dim: int = 96

	@nn.compact
	def __call__(self, input):
		output = layers.Conv(
			out_dim=self.out_dim,
			stride=2,
			)(input)
		output = GCViTDownsample()(output)
		return output


class GlobalQuery(nn.Module):
	"""
	Global query extractor.
	"""
	levels: int

	@nn.compact
	def __call__(self, input):
		reductions = self.levels
		for _ in range(max(1, self.levels)):
			input = GCViTFusedMBConv()(input)

			if reductions:
				input = layers.max_pool(input, stride=2)
				reductions -= 1

		return input


class WindowMHSA(nn.Module):
	"""
	Window multi-headed self-attention, with support for
	global queries.

	Args:
		n_heads (int): Number of heads.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		global_q (bool): Whether to use the provided global queries.
		Default is False.
	"""
	n_heads: int
	window_size: int
	global_q: bool = False

	@nn.compact
	def __call__(self, input, q_global=None):
		bs, h, w, in_dim = input.shape
		head_dim = in_dim//self.n_heads
		
		output = layers.window_partition(input, self.window_size)
		output = jnp.reshape(output, (-1, self.window_size**2, in_dim))

		if self.global_q:
			q = jnp.repeat(q_global, repeats=bs//len(q_global), axis=0)
			q = jnp.reshape(q, (bs, -1, self.n_heads, head_dim))
			q = jnp.swapaxes(q, axis1=1, axis2=2)

			kv = nn.Dense(
				features=2*in_dim,
				)(output)
			kv = jnp.reshape(kv, (-1, self.window_size**2, 2, self.n_heads, head_dim))
			kv = jnp.transpose(kv, (2, 0, 3, 1, 4))
			k, v = jnp.split(
				ary=kv,
				indices_or_sections=2,
				axis=0,
				)
			k, v = jnp.squeeze(k, axis=0), jnp.squeeze(v, axis=0)
		
		else:
			q, k, v = layers.QKV(
				n_heads=self.n_heads,
				)(output)

		output = layers.MHSA(
			pre_softmax=partial(
				layers.RelPosEmbed,
				n_heads=self.n_heads,
				window_size=self.window_size, 
				class_token=False,
				),
			)(q=q, k=k, v=v)
		output = layers.window_merge(
			output,
			img_size=(h, w),
			window_size=self.window_size,
			)

		return output


class GCViTBlock(nn.Module):
	"""
	GC ViT block.

	Args:
		n_heads (int): Number of heads.
		global_q (bool): Whether to use the provided global queries.
		Default is False.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion for the 
		hidden layer of the MLP.
		Default is 4.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
	"""
	n_heads: int
	global_q: bool = False
	mlp_hidden_dim_expansion_factor: int = 3
	window_size: int = 7
	layer_scale_init_value: T.Optional[float] = None

	@nn.compact
	def __call__(self, input, q_global=None):
		output = nn.LayerNorm(
			epsilon=1e-5,
			)(input)
		output = WindowMHSA(
			n_heads=self.n_heads,
			window_size=self.window_size,
			global_q=self.global_q,
			)(output, q_global=q_global)
		output = layers.LayerScale(
			init_value=self.layer_scale_init_value,
			)(output)
		output = input+output

		output = layers.TransformerMLP(
			hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
			layer_norm_eps=1e-5,
			layer_scale_init_value=self.layer_scale_init_value,
			)(output)

		return output


class GCViTStage(nn.Module):
	"""
	GC ViT stage.

	Args:
		depth (int): Depth.
		n_heads (int): Number of heads.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion for the 
		hidden layer of the MLP.
		Default is 4.
		window_size (int): Window size for relative position
		embedding and window attention. This value is used
		along both spatial dimensions.
		Default is 7.
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
		final_norm (bool): Whether to apply normalization on the output.
		Default is False.
		downsample (bool): Whether to downsample.
		Default is False.
	"""
	depth: int
	n_heads: int
	mlp_hidden_dim_expansion_factor: int = 3
	window_size: int = 7
	layer_scale_init_value: T.Optional[float] = None
	final_norm: bool = False
	downsample: bool = False

	@nn.compact
	def __call__(self, input):
		if self.downsample:
			input = GCViTDownsample(
				out_dim=2*input.shape[-1],
				)(input)
			
		q_global = GlobalQuery(
			levels=int(log2(input.shape[-2]//self.window_size)),
			)(input)
		for block_ind in range(self.depth):
			input = GCViTBlock(
				n_heads=self.n_heads,
				global_q=False if block_ind%2 == 0 else True,
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
				window_size=self.window_size,
				layer_scale_init_value=self.layer_scale_init_value,
				)(input, q_global)
			
		if self.final_norm:
			input = nn.LayerNorm(
				epsilon=1e-5,
				)(input)

		return input


class GCViT(nn.Module):
	"""
	Global context vision transformer.

	Args:
		depths (T.Tuple[int, ...]): Depth of each stage.
		token_dim (int): Token dimension.
		n_heads (T.Tuple[int, ...]): Number of heads of each
		stage.
		mlp_hidden_dim_expansion_factor (float): Factor of expansion for the 
		hidden layer of the MLP.
		Default is 4.
		window_size_factors (T.Tuple[int, ...]): Factors by which
		the spatial dimensions of each stage are divided to obtain the window
		size.
		Default is (32, 32, 16, 32).
		layer_scale_init_value (T.Optional[float]): Value
		for initializing LayerScale. If None, no LayerScale
		is applied.
		Default is None.
		n_classes (int): Number of output classes. If 0, there is no head,
		and the raw final features are returned. If -1, all stages of the 
		head, other than the final linear layer, are applied and the output
		returned.
		Default is 0.
	"""
	depths: T.Tuple[int, ...]
	token_dim: int
	n_heads: T.Tuple[int, ...]
	mlp_hidden_dim_expansion_factor: int = 3
	window_size_factors: T.Tuple[int, ...] = (32, 32, 16, 32)
	layer_scale_init_value: T.Optional[float] = None
	n_classes: int = 0

	@nn.compact
	def __call__(self, input):
		output = GCViTStem(
			out_dim=self.token_dim,
			)(input)
		self.sow(
			col='intermediates',
			name='stage_0',
			value=output,
			)

		for stage_ind in range(len(self.depths)):
			output = GCViTStage(
				depth=self.depths[stage_ind],
				n_heads=self.n_heads[stage_ind],
				mlp_hidden_dim_expansion_factor=self.mlp_hidden_dim_expansion_factor,
				window_size=input.shape[-2]//self.window_size_factors[stage_ind],
				layer_scale_init_value=self.layer_scale_init_value,
				final_norm=(stage_ind == len(self.depths)-1),
				downsample=False if stage_ind == 0 else True,
				)(output)
			self.sow(
				col='intermediates',
				name=f'stage_{stage_ind+1}',
				value=output,
				)
		
		output = layers.Head(
			n_classes=self.n_classes,
			)(output)

		return output


@register_configs
def get_gcvit_configs() -> T.Tuple[T.Type[GCViT], T.Dict]:
	"""
	Gets configurations for all available
	Swin models.

	Returns (T.Tuple[T.Type[GCViT], T.Dict]): The GCViT class and
	configurations of all available models.
	"""
	configs = {
		'gcvit_xxtiny_224': {
			'depths': (2, 2, 6, 2),
			'token_dim': 64,
			'n_heads': (2, 4, 8, 16),
			},
		'gcvit_xtiny_224': {
			'depths': (3, 4, 6, 5),
			'token_dim': 64,
			'n_heads': (2, 4, 8, 16),
			},
		'gcvit_tiny_224': {
			'depths': (3, 4, 19, 5),
			'token_dim': 64,
			'n_heads': (2, 4, 8, 16),
			},
		'gcvit_small_224': {
			'depths': (3, 4, 19, 5),
			'token_dim': 96,
			'n_heads': (3, 6, 12, 24),
			'mlp_hidden_dim_expansion_factor': 2,
			'layer_scale_init_value': 1e-5,
			},
		'gcvit_base_224': {
			'depths': (3, 4, 19, 5),
			'token_dim': 128,
			'n_heads': (4, 8, 16, 32),
			'mlp_hidden_dim_expansion_factor': 2,
			'layer_scale_init_value': 1e-5,
			},
		}
	return GCViT, configs



from setuptools import find_packages, setup


setup(
	name='flaim',
	version='0.0.2',
	description='Flax Image Models',
	long_description=open('README.md').read(),
	long_description_content_type='text/markdown',
	author='Borna Ahmadzadeh',
	author_email='borna.ahz@gmail.com',
	url='https://github.com/bobmcdear/flaim',
	packages=find_packages(),
	license='GNU',
	keywords=[
		'computer vision',
		'machine learning',
		'deep learning',
		'jax',
		'flax',
		],
	install_requires=[
		'jaxlib',
		'numpy>=1.20',
		'jax>=0.3.23',
		'flax>=0.6.0'
		],
	python_requires='>=3.8'
	)



"""
VGG.
"""


import typing as T

from flax import linen as nn

from .. import layers
from .factory import register_configs


class VGG(nn.Module):
	"""
	VGG network.

	Args:
		out_dims (T.Tuple[T.Union[int, str], ...]): Number of output channels of
		each convolution-ReLU block. If 'm', max pooling is used.
		bn (bool): Whether to have batch normalization after every convolution.
		If False, the training argument is ignored.
		Default is False.
		n_classes (int): Number of output classes. If 0, there is no head,
		head_conv_mlp_dim is ignored, and the raw final features are returned.
		If -1, all stages of the  head, other than the final linear layer, are applied
		and the output returned.
	"""
	out_dims: T.Tuple[T.Union[int, str], ...]
	bn: bool = False
	n_classes: int = 0

	@nn.compact
	def __call__(self, input, training: bool = True):
		stage_ind = 0
		for width in self.out_dims:
			if isinstance(width, int):
				input = layers.ConvBNAct(
					out_dim=width,
					bias_force=True,
					bn=self.bn,
					act=nn.relu,
					)(input, training=training)
			
			elif width == 'm':
				self.sow(
					col='intermediates',
					name=f'stage_{stage_ind}',
					value=input,
					)
				input = layers.max_pool(
					input=input,
					kernel_size=2,
					stride=2,
					padding=0,
					)
				stage_ind += 1
		
		output = layers.Head(
			n_classes=self.n_classes,
			)(input)
			
		return output


@register_configs
def get_vgg_configs() -> T.Tuple[T.Type[VGG], T.Dict]:
	"""
	Gets configurations for all available
	VGG models.

	Returns (T.Tuple[T.Type[VGG], T.Dict]): VGG class and
	configurations of all models.
	"""
	configs = {
		'vgg11': {
			'out_dims': (64, 'm', 128, 'm', 256, 256, 'm', 512, 512, 'm', 512, 512, 'm'),
			},
		'vgg13': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 'm', 512, 512, 'm', 512, 512, 'm'),
			},
		'vgg16': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 256, 'm', 512, 512, 512, 'm', 512, 512, 512, 'm'),
			},
		'vgg19': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 256, 256, 'm', 512, 512, 512, 512, 'm', 512, 512, 512, 512, 'm'),
			},
		'vgg11_bn': {
			'out_dims': (64, 'm', 128, 'm', 256, 256, 'm', 512, 512, 'm', 512, 512, 'm'),
			'bn': True,
			},
		'vgg13_bn': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 'm', 512, 512, 'm', 512, 512, 'm'),
			'bn': True,
			},
		'vgg16_bn': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 256, 'm', 512, 512, 512, 'm', 512, 512, 512, 'm'),
			'bn': True,
			},
		'vgg19_bn': {
			'out_dims': (64, 64, 'm', 128, 128, 'm', 256, 256, 256, 256, 'm', 512, 512, 512, 512, 'm', 512, 512, 512, 512, 'm'),
			'bn': True,
			},
		}
	return VGG, configs
